#"""
#Period simulation evaluator
#"""
#
#
##[
#from collections.abc import (Iterable, )
#from types import (EllipsisType, )
#from typing import (Any, )
#from numbers import (Number, )
#import itertools as _it
#import numpy as _np
#
#from .. import equations as _equations
#from .. import quantities as _quantities
#from ..equators import steady as _equators
#from ..jacobians import steady as _jacobians
#from ..models import variants as _variants
#
#from . import printers as _printers
##]
#
#
#class PeriodEvaluator:
#    """
#    """
#    #[
#    _equator_factory = _equators.PeriodEquator
#    _jacobian_factory = _jacobians.PeriodJacobian
#    _iter_printer_factory: = _printers.PeriodIterPrinter
#
#    def __init__(
#        self,
#        variant: _variants.Variant,
#        all_quantities: Iterable[_quantities.Quantity],
#        wrt_equations: Iterable[_equations.Equation],
#        wrt_qids: list[int],
#        data: _np.ndarray,
#        column: int,
#        /,
#        custom_functions: dict | None = None,
#        iter_printer_settings: dict[str, Any] | None = None,
#    ) -> None:
#        """
#        """
#        wrt_qids = list(wrt_qids, )
#        wrt_equations = list(wrt_equations, )
#        qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
#        qid_to_name = _quantities.create_qid_to_name(all_quantities, )
#        self._index_logly = list(_quantities.generate_index_logly(self.wrt_qids, qid_to_logly, ))
#
#        init_guess = 
#
#        maybelog_levels, maybelog_changes = variant.retrieve_maybelog_values_for_qids(self.wrt_qids, qid_to_logly, )
#        maybelog_levels, maybelog_changes = _fill_missing(maybelog_levels, maybelog_changes, )
#        self._maybelog_init_levels = maybelog_levels
#        self._maybelog_init_changes = maybelog_changes
#        #
#        self._num_levels = sum(self._bool_index_wrt_levels)
#        self._num_changes = sum(self._bool_index_wrt_changes)
#        #
#        self.init_guess = _np.hstack((
#            self._maybelog_init_levels[self._bool_index_wrt_levels],
#            self._maybelog_init_changes[self._bool_index_wrt_changes],
#        ))
#        #
#        self._steady_array = variant.create_steady_array(qid_to_logly, num_columns=shift_vec.shape[1], shift_in_first_column=-t_zero, )
#        self._update_steady_array(self.init_guess)
#        #
#        # Set up components
#        self._equator = self._equator_factory(
#            wrt_equations,
#            custom_functions=custom_functions,
#        )
#        self._jacobian = self._jacobian_factory(
#            wrt_equations,
#            self.wrt_qids,
#            qid_to_logly,
#            custom_functions=custom_functions,
#        )
#        iter_printer_settings = iter_printer_settings or {}
#        self.iter_printer = self._iter_printer_factory(
#            wrt_equations,
#            self.wrt_qids,
#            qid_to_logly,
#            qid_to_name,
#            **iter_printer_settings,
#        )
#
#    def eval(
#        self,
#        maybelog_guess: _np.ndarray,
#        /,
#    ) -> tuple[_np.ndarray, _np.ndarray]:
#        """
#        """
#        self._update_steady_array(maybelog_guess, )
#        equator = self._equator.eval(self._steady_array, )
#        jacobian = self._jacobian.eval(self._steady_array, )
#        jacobian = jacobian[:, self._bool_index_wrt_levels + self._bool_index_wrt_changes]
#        self.iter_printer.next(maybelog_guess, equator, True, )
#        return equator, jacobian
#
#    def _merge_levels_and_changes(
#        self,
#        wrt_qids_levels: list[str],
#        wrt_qids_changes: list[str],
#        /,
#    ) -> None:
#        """
#        """
#        ...
#
#    def _update_steady_array(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> None:
#        """
#        """
#        ...
#
#    def extract_levels(
#        self,
#        guess: _np.ndarray,
#        /,
#    ) -> tuple[_np.ndarray, tuple[int, ...]]:
#        """
#        """
#        levels = self._get_maybelog_levels(guess, )
#        levels[self._index_logly] = _np.exp(levels[self._index_logly])
#        levels = levels[self._bool_index_wrt_levels]
#        wrt_qids_levels = tuple(_it.compress(self.wrt_qids, self._bool_index_wrt_levels, ))
#        return levels, wrt_qids_levels
#
#    def extract_changes(
#        self,
#        guess: _np.ndarray,
#        /,
#    ) -> tuple[_np.ndarray, tuple[int, ...]]:
#        """
#        """
#        changes = self._get_maybelog_changes(guess, )
#        changes[self._index_logly] = _np.exp(changes[self._index_logly])
#        changes = changes[self._bool_index_wrt_changes]
#        wrt_qids_changes = tuple(_it.compress(self.wrt_qids, self._bool_index_wrt_changes, ))
#        return changes, wrt_qids_changes
#
#    def _get_maybelog_levels(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> _np.ndarray:
#        """
#        """
#        new_maybelog_levels = _np.copy(self._maybelog_init_levels)
#        if self._bool_index_wrt_levels:
#            new_maybelog_levels[self._bool_index_wrt_levels] = current_guess
#        return new_maybelog_levels
#    #]
#
#
#class FlatSteadyEvaluator(SteadyEvaluator, ):
#    """
#    """
#    #[
#    _equator_factory = _equators.FlatSteadyEquator
#    _jacobian_factory = _jacobians.FlatSteadyJacobian
#    _iter_printer_factory = _printers.FlatSteadyIterPrinter
#
#    def __init__(
#        self,
#        variant: _variants.Variant,
#        all_quantities: Iterable[_quantities.Quantity],
#        *args,
#        **kwargs,
#    ) -> None:
#        """
#        """
#        super().__init__(variant, all_quantities, *args, **kwargs, )
#        qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
#        variant.zero_changes(qid_to_logly, )
#
#    def _merge_levels_and_changes(
#        self,
#        wrt_qids_levels: list[int],
#        wrt_qids_changes: list[int],
#        /,
#    ) -> None:
#        self.wrt_qids = list(set(wrt_qids_levels))
#        self._bool_index_wrt_levels = [qid in wrt_qids_levels for qid in self.wrt_qids]
#        self._bool_index_wrt_changes = []
#
#    def _get_maybelog_levels(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> _np.ndarray:
#        """
#        """
#        new_maybelog_levels = _np.copy(self._maybelog_init_levels)
#        if self._bool_index_wrt_levels:
#            new_maybelog_levels[self._bool_index_wrt_levels] = current_guess
#        return new_maybelog_levels
#
#    def _get_maybelog_changes(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> _np.ndarray:
#        """
#        """
#        return _np.zeros(self._maybelog_init_changes.shape, dtype=float, )
#
#    def _update_steady_array(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> None:
#        """
#        """
#        new_maybelog_levels = self._get_maybelog_levels(current_guess, )
#        new_paths = _np.repeat(new_maybelog_levels.reshape(-1, 1), self._num_columns, axis=1, )
#        new_paths[self._index_logly, :] = _np.exp(new_paths[self._index_logly, :])
#        self._steady_array[self.wrt_qids, :] = new_paths
#    #]
#
#
#class NonflatSteadyEvaluator(SteadyEvaluator, ):
#    """
#    """
#    #[
#    _equator_factory = _NonflatSteadyEquator
#    _jacobian_factory = _NonflatSteadyJacobian
#    _iter_printer_factory = _printers.NonflatSteadyIterPrinter
#
#    def _merge_levels_and_changes(
#        self,
#        wrt_qids_levels: list[str],
#        wrt_qids_changes: list[str],
#        /,
#    ) -> None:
#        self.wrt_qids = list(set(wrt_qids_levels) | set(wrt_qids_changes))
#        self._bool_index_wrt_levels = [qid in wrt_qids_levels for qid in self.wrt_qids]
#        self._bool_index_wrt_changes = [qid in wrt_qids_changes for qid in self.wrt_qids]
#
#    def _get_maybelog_levels(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> _np.ndarray:
#        """
#        """
#        new_maybelog_levels = _np.copy(self._maybelog_init_levels)
#        if self._bool_index_wrt_levels:
#            new_maybelog_levels[self._bool_index_wrt_levels] = current_guess[:self._num_levels]
#        return new_maybelog_levels
#
#    def _get_maybelog_changes(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> _np.ndarray:
#        """
#        """
#        new_maybelog_changes = _np.copy(self._maybelog_init_changes)
#        if self._bool_index_wrt_changes:
#            new_maybelog_changes[self._bool_index_wrt_changes] = current_guess[self._num_levels:]
#        return new_maybelog_changes
#
#    def _update_steady_array(
#        self,
#        current_guess: _np.ndarray,
#        /,
#    ) -> None:
#        """
#        """
#        new_maybelog_levels = self._get_maybelog_levels(current_guess, )
#        new_maybelog_changes = self._get_maybelog_changes(current_guess, )
#        new_paths = new_maybelog_levels.reshape(-1, 1) + self._shift_vec * new_maybelog_changes.reshape(-1, 1)
#        new_paths[self._index_logly, :] = _np.exp(new_paths[self._index_logly, :])
#        self._steady_array[self.wrt_qids, :] = new_paths
#    #]
#
#
#def _fill_missing(
#    maybelog_levels: _np.ndarray,
#    maybelog_changes: _np.ndarray,
#    /,
#    default_level: Number = 1/9,
#    default_change: Number = 0,
#) -> tuple[_np.ndarray, _np.ndarray]:
#    """
#    """
#    #[
#    maybelog_levels[_np.isnan(maybelog_levels)] = default_level
#    maybelog_changes[_np.isnan(maybelog_changes)] = default_change
#    return maybelog_levels, maybelog_changes
#    #]
#
#
#def _prepare_time_shifts(
#    equations: Iterable[_equations.Equation],
#    /,
#) -> tuple[_np.ndarray, int]:
#    """
#    Return [min_shift, ..., max_shift] and the index of the column with shift=0
#    """
#    #[
#    # Add 1 to max shift to accommodate nonflat steady equators
#    add_shift = 1
#    #
#    min_shift = _equations.get_min_shift_from_equations(equations, )
#    max_shift = _equations.get_max_shift_from_equations(equations, ) + add_shift
#    num_columns = -min_shift + 1 + max_shift
#    shift_in_first_column = min_shift
#    shift_vec = _np.array(range(min_shift, max_shift+1, ), dtype=float, ).reshape(1, -1, )
#    t_zero = -min_shift
#    return shift_vec, t_zero
#    #]
#
#



"""
"""


#[
from __future__ import annotations

from typing import (Any, )
from collections.abc import (Iterable, )

from .. import quantities as _quantities
from ..databoxes import main as _databoxes
#]


class Variant:
    """
    """
    #[

    __slots__ = (
        "parameters",
    )

    def __init__(
        self,
        quantities: Iterable[_quantities.Quantity] | None = None,
        /,
    ) -> None:
        """
        """
        self.parameters = {
            qty.human: None
            for qty in quantities
            if qty.kind is _quantities.QuantityKind.PARAMETER
        }

    def assign(self, *args, **kwargs, ) -> None:
        """
        """
        for a in args:
            self._assign_from_kwargs(**a, )
        self._assign_from_kwargs(**kwargs, )

    def _assign_from_kwargs(self, **kwargs, ) -> None:
        """
        """
        existing_keys = set(self.parameters.keys())
        new_keys = set(kwargs.keys())
        for n in existing_keys & new_keys:
            self.parameters[n] = kwargs[n]

    #]


class VariantMixin:
    """
    """
    #[

    def assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        """
        for v in self._variants:
            v.assign(*args, **kwargs, )

    def get_parameters(self, /, ) -> _databoxes.Databox:
        """
        """
        parameters = self._variants[0].parameters
        return _databoxes.Databox.from_dict(parameters, )

    #]




"""
Dynamic function makers
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Any, Callable, )
import copy as _cp

from .aldi import adaptations as _adaptations
#]


def make_lambda(
    args: Iterable[str],
    expression: str,
    globals: dict[str, Any] | None = None,
) -> tuple[Callable, str, dict[str, Any]]:
    """
    Create lambda from the list of args and an expression
    """
    #[
    func_str = f"lambda {', '.join(args, )}: {str(expression)}"
    func = eval(func_str, globals, )
    return func, func_str, globals
    #]


def prepare_globals(
    *,
    custom_functions: dict[str, Any] | None = None,
    builtins: dict[str, Callable] | None = None,
) -> dict[str, Any]:
    """
    Create globals dict restricting builtins by default
    """
    #[
    globals_ = _cp.deepcopy(custom_functions, ) if custom_functions else {}
    globals_["__builtins__"] = builtins
    globals_ = _adaptations.add_function_adaptations_to_custom_functions(globals_, )
    return globals_
    #]




"""
Data arrays with row and column names
"""


#[
from __future__ import annotations

from typing import (Self, Protocol, )
from numbers import (Number, )
from collections.abc import (Iterable, )
import numpy as _np
import dataclasses as _dc
import numpy as _np

from .series import main as _series
from .databoxes import main as _databoxes
from .plans import main as _plans
from .incidences import main as _incidences
from . import dates as _dates
#]


__all__ = (
    "HorizontalDataslate",
    "VerticalDataslate",
)


class SlatableProtocol(Protocol, ):
    """
    """
    def get_min_max_shift(self, /, ) -> tuple[int, int]:  ...
    def get_databox_names(self, /, ) -> tuple[str, ...]:  ...


@_dc.dataclass
class _Dataslate:
    """
    """
    #[

    data: _np.ndarray | None = None
    missing_names: tuple[str, ...] | None = None
    base_indices: tuple[int, ...] | None = None
    _names: Iterable[str] | None = None
    _descriptions: Iterable[str] | None = None
    _dates: tuple[_dates.Dater, ...] | None = None
    _record_shape: tuple[int, int] | None = None

    @classmethod
    def for_slatable(
        klass,
        slatable: SlatableProtocol,
        databox: _databoxes.Databox,
        base_range: Iterable[_dates.Dater],
        /,
        variant: int = 0,
        plan: _plans.Plan | None = None,
    ) -> Self:
        """
        """
        self = klass()
        self._names = tuple(slatable.get_databox_names(plan, ))
        self.missing_names = tuple(databox.get_missing_names(self._names, ))
        self._resolve_dates(slatable, base_range, )
        self._populate_data(databox, variant, )
        self._populate_descriptions(databox, )
        return self

    @classmethod
    def from_databox(
        klass,
        databox: _databoxes.Databox,
        names: Iterable[str],
        base_range: Iterable[_dates.Dater],
        /,
        variant: int = 0,
    ) -> Self:
        """
        """
        self = klass()
        self._names = tuple(names)
        self.missing_names = tuple(databox.get_missing_names(self._names, ))
        self._dates = tuple(base_range)
        self._base_indices = tuple(range(len(self._dates)))
        self._populate_data(databox, variant, )
        self._populate_descriptions(databox, )
        return self

    def _populate_data(
        self,
        databox: _databoxes.Databox,
        variant: int,
    ) -> None:
        """
        """
        data = tuple(
            _extract_data_from_record(databox[n], self.from_to, self.num_periods, variant, ).reshape(*self._record_reshape, )
            if n not in self.missing_names else self.nan_row.reshape(*self._record_reshape, )
            for n in self._names
        )
        self.data = self._stack_in(data, )

    def _populate_descriptions(
        self,
        databox: _databoxes.Databox,
        /,
    ) -> None:
        """
        """
        self._descriptions = tuple(
            _extract_descriptions_from_record(databox[n], )
            if n not in self.missing_names else ""
            for n in self._names
        )

    def to_databox(self, *args, **kwargs, ) -> _dates.Databox:
        """
        """
        return multiple_to_databox((self,), *args, **kwargs, )

    @property
    def num_periods(self, /, ) -> int:
        """
        """
        return len(self._dates)

    @property
    def from_to(self, /, ) -> tuple[_dates.Dater, _dates.Dater]:
        """
        """
        return self._dates[0], self._dates[-1]

    @property
    def nan_row(self, /, ) -> int:
        """
        """
        return _np.full((self.num_periods, ), _np.nan, dtype=float)

    @property
    def num_rows(self, /, ) -> int:
        """
        """
        return self.data.shape[0] if self.data is not None else 0

    @property
    def base_slice(self, /, ) -> slice:
        """
        """
        return slice(self._base_indices[0], self._base_indices[-1]+1)

    def remove_terminal(self, /, ) -> None:
        """
        """
        last_base_index = self._base_indices[-1]
        self.data = self.data[:, :last_base_index+1]
        self._dates = self._dates[:last_base_index+1]

    def copy_data(self, /, ) -> _np.ndarray:
        """
        """
        return self.data.copy()

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return { name: row for row, name in enumerate(self._names, ) }

    def _resolve_dates(
        self,
        slatable: SlatableProtocol,
        base_range: Iterable[_dates.Dater],
        /,
    ) -> None:
        self._dates, self._base_indices = get_extended_range(slatable, base_range, )

    @staticmethod
    def retrieve_vector_from_data_array(
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
        /,
    ) -> _np.ndarray:
        ...

    def retrieve_vector(
        self,
        tokens: tuple[str, ...],
        index_zero: int,
        /,
    ) -> _np.ndarray:
        """
        """
        return self.retrieve_vector_from_data_array(self.data, tokens, index_zero, )

    @staticmethod
    def store_vector_in_data_array(
        vector: _np.ndarray,
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
        /,
    ) -> None:
        ...

    def store_vector(
        self,
        tokens: tuple[str, ...],
        vector: _np.ndarray,
        index_zero: int,
        /,
    ) -> None:
        """
        """
        store_vector_in_horizontal_data_array(vector, self.data, tokens, index_zero, )

    def retrieve_record(
        self,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> _np.ndarray:
        """
        """
        ...

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> None:
        """
        """
        ...

    def fill_missing_in_base_periods(
        self,
        names,
        /,
        fill: Number = 0,
    ) -> None:
        """
        """
        for i, n in enumerate(self._names):
            if names is not Ellipsis and n not in names:
                continue
            values = self.retrieve_record(i, self._base_indices, )
            values[_np.isnan(values)] = fill
            self.store_record(values, i, self._base_indices, )

    #]


class HorizontalDataslate(_Dataslate, ):
    """
    """
    #[
    _record_reshape = (1, -1, )

    @staticmethod
    def _stack_in(data, /, ) -> _np.ndarray:
        """
        """
        return _np.vstack(data)

    @staticmethod
    def retrieve_vector_from_data_array(
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
        /,
    ) -> _np.ndarray:
        """
        """
        rows, columns = _incidences.rows_and_columns_from_tokens_in_horizontal(tokens, index_zero, )
        return data[rows, columns].reshape(-1, 1)

    @staticmethod
    def store_vector_in_data_array(
        vector: _np.ndarray,
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
    ) -> None:
        """
        """
        rows, columns = _incidences.rows_and_columns_from_tokens_in_horizontal(tokens, index_zero, )
        data[rows, columns] = vector

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove < 0:
            raise ValueError("Cannot remove negative number of columns from start")
        if remove:
            self.data = self.data[:, remove:]
            self._dates = self._dates[remove:]

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove > 0:
            raise ValueError("Cannot remove positive number of columns from end")
        if remove:
            self.data = self.data[:, :remove]
            self._dates = self._dates[:remove]

    @property
    def column_dates(self, /, ) -> tuple[_dates.Dater, ...]:
        """
        """
        return self._dates

    @property
    def base_columns(self, /, ) -> tuple[int, ...]:
        """
        """
        return self._base_indices

    @property
    def row_names(self, /, ) -> Iterable[str]:
        """
        """
        return self._names

    def retrieve_record(
        self,
        record_id: int,
        /,
        columns: slice | Iterable[int] | None = None,
    ) -> _np.ndarray:
        """
        """
        if columns is None:
            return self.data[record_id, :]
        else:
            return self.data[record_id, columns]

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        columns: slice | Iterable[int] | None = None,
    ) -> None:
        """
        """
        if columns is None:
            self.data[record_id, :] = values
        else:
            self.data[record_id, columns] = values

    #]


class VerticalDataslate(_Dataslate, ):
    """
    """
    #[
    _record_reshape = (-1, 1, )

    @staticmethod
    def _stack_in(data, /, ) -> _np.ndarray:
        """
        """
        return _np.hstack(data)

    @property
    def row_dates(self, /, ) -> tuple[_dates.Dater, ...]:
        """
        """
        return self._dates

    @property
    def column_names(self, /, ) -> tuple[str]:
        """
        """
        return self._names

    def retrieve_record(
        self,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> _np.ndarray:
        """
        """
        if slice_ is None:
            return self.data[:, record_id]
        else:
            return self.data[slice_, record_id]

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> None:
        """
        """
        if slice_ is None:
            self.data[:, record_id] = values
        else:
            self.data[slice_, record_id] = values

    #]


def multiple_to_databox(
    slates,
    /,
    target_databox: _databoxes.Databox | None = None,
) -> _databoxes.Databox:
    """
    Add data from a dataslate to a new or existing databox
    """
    #[
    if target_databox is None:
        target_databox = _databoxes.Databox()
    num_columns = len(slates)
    start_date = slates[0]._dates[0]
    for record_id, n in enumerate(slates[0]._names):
        data = _np.hstack(tuple(
            ds.retrieve_record(record_id, ).reshape(-1, 1)
            for ds in slates
        ))
        target_databox[n] = _series.Series(
            num_columns=num_columns,
            start_date=start_date,
            values=data,
            description=slates[0]._descriptions[record_id],
        )
    return target_databox
    #]


def get_extended_range(
    slatable: SlatableProtocol,
    base_range: Iterable[_dates.Dater],
    /,
) -> Iterable[_dates.Dater]:
    """
    """
    base_range = tuple(t for t in base_range)
    num_base_periods = len(base_range)
    min_shift, max_shift = slatable.get_min_max_shifts()
    if min_shift == 0:
        min_shift = -1
    min_base_date = min(base_range)
    max_base_date = max(base_range)
    start_date = min_base_date + min_shift
    end_date = max_base_date + max_shift
    base_indices = tuple(_dates.date_index(base_range, start_date))
    # base_indices = tuple(range(-min_shift, -min_shift+num_base_periods))
    extended_dates = tuple(_dates.Ranger(start_date, end_date))
    return extended_dates, base_indices


def _extract_data_from_record(record, from_to, num_periods, column, /, ):
    """
    """
    try:
        # Record is a time series
        return record.get_data_column_from_to(from_to, column)
    except AttributeError:
        # Record is a numeric scalar
        return _np.full((num_periods, ), float(record), dtype=float, )


def _extract_descriptions_from_record(record, /, ):
    """
    """
    try:
        # Record is a time series
        return str(record.get_description())
    except AttributeError:
        # Record is a numeric scalar
        return ""
    except ValueError:
        # Record is a numeric scalar
        return ""




"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np

from ..databoxes import main as _databoxes
from ..fords import simulators as _simulators
from ..plans import main as _plans
from .. import dataslates as _dataslates
#]


class SimulateMixin:
    """
    """
    #[

    def simulate(
        self,
        in_databox: _databoxes.Databox,
        base_range: Iterable[Dater],
        /,
        anticipate: bool = True,
        deviation: bool = False,
        plan: _plans.Plan | None = None,
        prepend_input: bool = True,
        target_databox: _databoxes.Databox | None = None,
    ) -> tuple[_databoxes.Databox, dict[str, Any]]:
        """
        """
        dataslates = tuple(
            _dataslates.HorizontalDataslate.for_slatable(self, in_databox, base_range, plan=plan, variant=i, )
            for i in range(self.num_variants)
        )
        #
        qid_to_logly = self.create_qid_to_logly()
        boolex_logly = tuple(
            qid_to_logly[qid] or False
            for qid in range(dataslates[0].num_rows)
        )
        #
        for variant, dataslate in zip(self._variants, dataslates):
            new_data = dataslate.copy_data()
            new_data = _simulators.simulate_flat(
                variant.solution, self.get_solution_vectors(), boolex_logly,
                new_data, dataslate.base_columns, deviation, anticipate,
            )
            dataslate.data = new_data
            dataslate.remove_periods_from_end(dataslate.base_columns[-1] - dataslate.num_periods + 1)
        #
        # Combine resulting dataslates into a databox
        #
        out_db = _dataslates.multiple_to_databox(dataslates, )
        if prepend_input:
            out_db.prepend(in_databox, base_range[0]-1, )
        out_db = out_db | self.get_parameters_stds()
        #
        # Add to custom databox
        #
        if target_databox is not None:
            out_db = target_databox | out_db
        #
        info = {}
        return out_db, info

    #]


